import hashlib
import json
import os
import threading
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Dict

from metaflow.exception import MetaflowException
from metaflow.metaflow_config import FAST_BAKERY_URL, get_pinned_conda_libs
from metaflow.metaflow_environment import MetaflowEnvironment
from metaflow.plugins.aws.batch.batch_decorator import BatchDecorator
from metaflow.plugins.kubernetes.kubernetes_decorator import KubernetesDecorator
from metaflow.plugins.pypi.conda_decorator import CondaStepDecorator
from metaflow.plugins.pypi.conda_environment import CondaEnvironment
from metaflow.plugins.pypi.pypi_decorator import PyPIStepDecorator
from metaflow import decorators

from .fast_bakery import FastBakery, FastBakeryApiResponse, FastBakeryException

BAKERY_METAFILE = ".imagebakery-cache"

import fcntl
import json
import os
from concurrent.futures import ThreadPoolExecutor
from functools import wraps


def cache_request(cache_file):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            call_args = kwargs.copy()
            call_args.update(zip(func.__code__.co_varnames, args))
            call_args.pop("self", None)
            call_args.pop("ref", None)
            # invalidate cache when moving from one deployment to another
            call_args.update({"fast_bakery_url": FAST_BAKERY_URL})
            cache_key = hashlib.md5(
                json.dumps(call_args, sort_keys=True).encode("utf-8")
            ).hexdigest()

            try:
                with open(cache_file, "r") as f:
                    cache = json.load(f)
                    if cache_key in cache:
                        return FastBakeryApiResponse(cache[cache_key])
            except (FileNotFoundError, json.JSONDecodeError):
                cache = {}

            result = func(*args, **kwargs)

            try:
                with open(cache_file, "r+") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    try:
                        f.seek(0)
                        cache = json.load(f)
                    except json.JSONDecodeError:
                        cache = {}

                    cache[cache_key] = result.response

                    f.seek(0)
                    f.truncate()
                    json.dump(cache, f)
            except FileNotFoundError:
                # path to cachefile might not exist.
                os.makedirs(os.path.dirname(cache_file), exist_ok=True)
                with open(cache_file, "w") as f:
                    fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                    json.dump({cache_key: result.response}, f)

            return result

        return wrapper

    return decorator


class DockerEnvironmentException(MetaflowException):
    headline = "Ran into an error while baking image"

    def __init__(self, msg):
        super(DockerEnvironmentException, self).__init__(msg)


class DockerEnvironment(MetaflowEnvironment):
    TYPE = "fast-bakery"
    _filecache = None

    def __init__(self, flow):
        self.skipped_steps = set()
        self.flow = flow

        self.results = {}
        self.images_baked = 0

    def set_local_root(self, local_root):
        self.local_root = local_root

    def decospecs(self):
        # Due to conflicts with the CondaEnvironment fallback and bakery,
        # we can not simply attach 'conda' or 'pypi' to all steps here.
        # Instead we do this on a per-step basis in init_environment
        return ("fast_bakery_internal",) + super().decospecs()

    def validate_environment(self, logger, datastore_type):
        self.datastore_type = datastore_type
        self.logger = logger

        # Avoiding circular imports.
        from metaflow.plugins import DATASTORES

        self.datastore = [d for d in DATASTORES if d.TYPE == self.datastore_type][0]

    def init_environment(self, echo):
        self.skipped_steps = {
            step.name for step in self.flow if not _step_executes_remotely(step)
        }
        # Attach environment decorator as needed. This is done on a step-by-step basis
        # as we require a conda decorator for fallback steps, but prefer pypi for the baked ones.
        for step in self.flow:
            # Mixing @pypi/@conda in a single step is not supported yet.
            # We validate this before attaching any new ones as the OSS Conda environment requires an implicit conda decorator for pypi environments which would fail the validation.
            if sum(1 for deco in step.decorators if _is_env_deco(deco)) > 1:
                raise MetaflowException(
                    "Mixing and matching PyPI packages and Conda packages within a\n"
                    "step is not yet supported. Use one of @pypi or @conda only for the *%s* step."
                    % step.name
                )
            if step.name in self.skipped_steps:
                # Conda fallback requires a conda decorator as the default for a step
                decorators._attach_decorators_to_step(step, ["conda"])
            else:
                if not _step_has_environment_deco(step):
                    # We default to PyPI for steps that are going to be baked.
                    decorators._attach_decorators_to_step(step, ["pypi"])
                    # init the attached decorator
            # Initialize the decorator we attached.
            # This is crucial for the conda decorator to work properly in the fallback environment
            decorators._init(self.flow)
            for deco in step.decorators:
                if _is_env_deco(deco):
                    deco.step_init(
                        self.flow,
                        None,  # not passing graph as it is not available, and not required by conda/pypi decorators
                        step.name,
                        step.decorators,
                        self,
                        self.datastore,
                        echo,
                    )

        steps_to_bake = [
            step
            for step in self.flow
            if step.name not in self.skipped_steps and not self.is_disabled(step)
        ]
        if steps_to_bake:
            self.logger("🚀 Baking container image(s) ...")
            start_time = time.time()
            self.results = self._bake(steps_to_bake)
            for step in steps_to_bake:
                for d in step.decorators:
                    if _is_remote_deco(d):
                        d.attributes["image"] = self.results[step.name].container_image
                        d.attributes["executable"] = self.results[step.name].python_path
            if self.images_baked > 0:
                bake_time = time.time() - start_time
                self.logger(
                    f"🎉 All container image(s) baked in {bake_time:.2f} seconds!"
                )
            else:
                self.logger("🎉 All container image(s) baked!")

        if self.skipped_steps:
            self.delegate = CondaEnvironment(self.flow)
            self.delegate.set_local_root(self.local_root)
            self.delegate.validate_environment(echo, self.datastore_type)
            self.delegate.init_environment(echo, self.skipped_steps)

    def _bake(self, steps) -> Dict[str, FastBakeryApiResponse]:
        metafile_path = get_fastbakery_metafile_path(self.local_root, self.flow.name)
        logger_lock = threading.Lock()

        @cache_request(metafile_path)
        def _cached_bake(
            ref=None,
            python=None,
            pypi_packages=None,
            conda_packages=None,
            base_image=None,
        ):
            try:
                bakery = FastBakery(url=FAST_BAKERY_URL)
                bakery._reset_payload()
                bakery.python_version(python)
                bakery.pypi_packages(pypi_packages)
                bakery.conda_packages(conda_packages)
                bakery.base_image(base_image)
                # bakery.ignore_cache()

                with logger_lock:
                    self.logger(f"🍳 Baking [{ref}] ...")
                    self.logger(f"     🐍 Python: {python}")

                    if pypi_packages:
                        self.logger(f"     📦 PyPI packages:")
                        for package, version in pypi_packages.items():
                            self.logger(f"        🔧 {package}: {version}")

                    if conda_packages:
                        self.logger(f"     📦 Conda packages:")
                        for package, version in conda_packages.items():
                            self.logger(f"        🔧 {package}: {version}")

                    self.logger(f"     🏗️  Base image: {base_image}")

                start_time = time.time()
                res = bakery.bake()
                # TODO: Get actual bake time from bakery
                bake_time = time.time() - start_time

                with logger_lock:
                    self.logger(f"🏁 Baked [{ref}] in {bake_time:.2f} seconds!")
                self.images_baked += 1
                return res
            except FastBakeryException as ex:
                raise DockerEnvironmentException(f"Bake [{ref}] failed: {str(ex)}")

        def prepare_step(step):
            base_image = next(
                (
                    d.attributes.get("image")
                    for d in step.decorators
                    if isinstance(d, (KubernetesDecorator))
                ),
                None,
            )
            dependencies = next(
                (d for d in step.decorators if _is_env_deco(d)),
                None,
            )
            python = next(
                (
                    d.attributes["python"]
                    for d in step.decorators
                    if isinstance(d, CondaStepDecorator)
                ),
                None,
            )
            pypi_deco = next(
                (d for d in step.decorators if isinstance(d, PyPIStepDecorator)), None
            )
            # if pypi decorator is set and user has specified a python version, we must create a new environment.
            # otherwise rely on the base environment
            if pypi_deco is not None:
                python = (
                    pypi_deco.attributes["python"]
                    if pypi_deco.is_attribute_user_defined("python")
                    else None
                )

            packages = get_pinned_conda_libs(python, self.datastore_type)
            packages.update(dependencies.attributes["packages"] if dependencies else {})

            requested = {
                "python": python,
                "pypi_packages": (
                    packages if isinstance(dependencies, PyPIStepDecorator) else None
                ),
                "conda_packages": (
                    packages if isinstance(dependencies, CondaStepDecorator) else None
                ),
                "base_image": base_image,
            }
            dedup_key = hashlib.sha256(
                json.dumps(requested).encode("utf-8")
            ).hexdigest()

            return step.name, dedup_key, requested

        with ThreadPoolExecutor() as executor:
            prepared_args = list(executor.map(prepare_step, steps))
            # Deduplicate the requests for baking images of steps.
            # We do not want to bake the same image twice.
            dedup_requests = {}
            for step_name, key, args in prepared_args:
                if key not in dedup_requests:
                    dedup_requests[key] = {"step_names": set(), "args": args}
                dedup_requests[key]["step_names"].add(step_name)

            # unique futures
            futures = []
            for i, kv in enumerate(dedup_requests.items(), 1):
                key, value = kv
                future = executor.submit(
                    _cached_bake, **{**value["args"], "ref": f"#{i:02d}"}
                )
                futures.append({"step_names": value["step_names"], "future": future})

            results = {}
            for item in futures:
                for step_name in item["step_names"]:
                    results[step_name] = item["future"].result()

            return results

    def executable(self, step_name, default=None):
        if step_name in self.skipped_steps:
            return self.delegate.executable(step_name, default)
        # default is set to the right executable
        if default is not None:
            return default
        if default is None and step_name in self.results:
            # try to read pythonpath from results. This can happen immediately after baking.
            return self.results[step_name].python_path
        # we lack a default and baking results. fallback to parent executable.
        return super().executable(step_name, default)

    def interpreter(self, step_name):
        if step_name in self.skipped_steps:
            return self.delegate.interpreter(step_name)
        return None

    def is_disabled(self, step):
        for decorator in step.decorators:
            # @conda decorator is guaranteed to exist thanks to self.decospecs
            if decorator.name in ["conda", "pypi"]:
                # handle @conda/@pypi(disabled=True)
                disabled = decorator.attributes["disabled"]
                return str(disabled).lower() == "true"
        return False

    def pylint_config(self):
        config = super().pylint_config()
        # Disable (import-error) in pylint
        config.append("--disable=F0401")
        return config

    def get_package_commands(self, codepackage_url, datastore_type):
        # we must set the skip install flag at this stage in order to skip package downloads,
        # doing so in bootstrap_commands is too late in the lifecycle.
        return [
            "export METAFLOW_SKIP_INSTALL_DEPENDENCIES=$FASTBAKERY_IMAGE",
        ] + super().get_package_commands(codepackage_url, datastore_type)

    def bootstrap_commands(self, step_name, datastore_type):
        if step_name in self.skipped_steps:
            return self.delegate.bootstrap_commands(step_name, datastore_type)
        return super().bootstrap_commands(step_name, datastore_type)


def get_fastbakery_metafile_path(local_root, flow_name):
    return os.path.join(local_root, flow_name, BAKERY_METAFILE)


def _is_remote_deco(deco):
    return isinstance(deco, (BatchDecorator, KubernetesDecorator))


def _step_executes_remotely(step):
    "Check if a step is going to execute remotely or locally"
    return any(_is_remote_deco(deco) for deco in step.decorators)


def _is_env_deco(deco):
    "Check if a decorator is a known environment decorator (conda/pypi)"
    return isinstance(deco, (PyPIStepDecorator, CondaStepDecorator))


def _step_has_environment_deco(step):
    "Check if a step has a virtual environment decorator"
    return any(_is_env_deco(deco) for deco in step.decorators)



from metaflow.exception import MetaflowException
from .constants import SUPPORTABLE_GPU_TYPES


class NvcfJobFailedException(MetaflowException):
    headline = "[@nvidia] error"

    def __init__(self, msg):
        super(NvcfJobFailedException, self).__init__(msg)


class NvcfPollingConnectionError(MetaflowException):
    headline = "[@nvidia] polling error."

    def __init__(self, og_error_msg):
        msg = (
            "An error occurred while polling the job status. "
            "\n\nOriginal error message: %s" % (og_error_msg)
        )

        super(NvcfPollingConnectionError, self).__init__(msg)


class RequestedGPUTypeUnavailableException(MetaflowException):
    headline = "[@nvidia RequestedGPUTypeUnavailableException] GPU type unavailable."

    def __init__(self, requested_gpu_type):
        msg = (
            f"The requested GPU type @nvidia(..., gpu_type='{requested_gpu_type}') is not available. "
            f"Please choose from the following supported GPU types when using @nvidia: {SUPPORTABLE_GPU_TYPES}"
        )
        super(RequestedGPUTypeUnavailableException, self).__init__(msg)


class UnsupportedNvcfConfigurationException(MetaflowException):
    headline = (
        "[@nvidia UnsupportedNvcfConfigurationException] Unsupported GPU configuration"
    )

    def __init__(self, n_gpu, gpu_type, available_configurations, step):
        msg = f"The requested configuration of @nvidia(gpu={n_gpu}, gpu_type='{gpu_type}') for @step {step} is not available."
        if len(available_configurations) == 0:
            msg += (
                "\n\nNo configurations are available in your Outerbounds deployment."
                " Please contact Outerbounds support if you wish to use @nvidia."
            )
        else:
            msg += f"\n\nAvailable configurations for your deployment include: \n\t- {self._display(available_configurations)}"
            msg += "\n\nPlease contact Outerbounds support if you wish to use a configuration not listed above."
        super(UnsupportedNvcfConfigurationException, self).__init__(msg)

    def _display(self, configs):
        _available_decos = []
        for cfg in configs:
            n_gpu, gpu_type = cfg[0], cfg[1]
            _available_decos.append(f"@nvidia(gpu={n_gpu}, gpu_type='{gpu_type}')")
        return "\n\t- ".join(_available_decos)


class UnsupportedNvcfDatastoreException(MetaflowException):
    headline = "[@nvidia UnsupportedNvcfDatastoreException] Unsupported datastore"

    def __init__(self, ds_type):
        msg = (
            "The *@nvidia* decorator requires --datastore=s3 or --datastore=azure or --datastore=gs at the moment."
            f"Current datastore type: {ds_type}."
        )
        super(UnsupportedNvcfDatastoreException, self).__init__(msg)


class NvcfTimeoutTooShortException(MetaflowException):
    headline = "[@nvidia NvcfTimeoutTooShortException] Timeout too short"

    def __init__(self, step):
        msg = (
            "The timeout for step *{step}* should be at least 60 seconds for "
            "execution with @nvidia".format(step=step)
        )
        super(NvcfTimeoutTooShortException, self).__init__(msg)


class NvcfQueueTimeoutTooShortException(MetaflowException):
    headline = "[@nvidia NvcfQueueTimeoutTooShortException] Queue Timeout too short"

    def __init__(self, step):
        msg = (
            "The queue timeout for step *{step}* should be at least 60 seconds for "
            "execution with @nvidia".format(step=step)
        )
        super(NvcfQueueTimeoutTooShortException, self).__init__(msg)


class NvcfKilledException(MetaflowException):
    headline = "Nvidia job killed"



from datetime import datetime
from metaflow.decorators import StepDecorator
from ...profilers.gpu import GPUProfiler  # Fix import
from .deco_injector import CardDecoratorInjector
import threading


class GPUProfileDecorator(StepDecorator):
    name = "gpu_profile"

    defaults = {
        "include_artifacts": True,
        "artifact_prefix": "gpu_profile_",
        "interval": 1,
    }

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        self.deco_injector = CardDecoratorInjector()
        self.deco_injector.attach_card_decorator(
            flow,
            step_name,
            "gpu_profile",
            "blank",
            refresh_interval=self.attributes["interval"],
        )

    def task_pre_step(
        self,
        step_name,
        task_datastore,
        metadata,
        run_id,
        task_id,
        flow,
        graph,
        retry_count,
        max_user_code_retries,
        ubf_context,
        inputs,
    ):
        self._profiler = GPUProfiler(
            interval=self.attributes["interval"],
            artifact_name=self.attributes["artifact_prefix"] + "data",
        )

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        from metaflow import current
        from metaflow.cards import Markdown

        if self.attributes["include_artifacts"]:
            setattr(
                flow,
                self.attributes["artifact_prefix"] + "num_gpus",
                len(self._profiler.devices),
            )

        current.card["gpu_profile"].append(
            Markdown("# GPU profile for `%s`" % current.pathspec)
        )
        current.card["gpu_profile"].append(
            Markdown(
                "_Started at: %s_"
                % datetime.now().astimezone().strftime("%Y-%m-%dT%H:%M:%S %z")
            )
        )
        self._profiler._setup_card()
        current.card["gpu_profile"].refresh()
        self._update_thread = threading.Thread(
            target=self._profiler._update_card, daemon=True
        )
        self._update_thread.start()

        def wrapped_step_func():
            try:
                step_func()
            finally:
                try:
                    results = self._profiler.finish()
                except:
                    results = {"error": "couldn't read profiler results"}
                if self.attributes["include_artifacts"]:
                    setattr(flow, self.attributes["artifact_prefix"] + "data", results)

        return wrapped_step_func



from metaflow.exception import MetaflowException
import os
from metaflow.metaflow_config_funcs import init_config
import requests
import time
import random

# IMPORTANT: Currently contents of this file are mostly duplicated from the outerbounds package.
# This is purely due to the time rush of having to deliver this feature. As a fast forward, we
# will reorganize things in a way that the amount of duplication in minimum.


APP_READY_POLL_TIMEOUT_SECONDS = 300
# Even after our backend validates that the app routes are ready, it takes a few seconds for
# the app to be accessible via the browser. Till we hunt down this delay, add an extra buffer.
APP_READY_EXTRA_BUFFER_SECONDS = 30


def start_app(port=-1, name=""):
    """
    Starts an app on the workstation.
    List workstations, looks for "NamedPorts", then makes an update call to the NamedPorts for the workstation.
    """
    if len(name) == 0 or len(name) >= 20:
        raise MetaflowException("App name should not be more than 20 characters long.")
    elif not name.isalnum() or not name.islower():
        raise MetaflowException(
            "App name can only contain lowercase alphanumeric characters."
        )

    if "WORKSTATION_ID" not in os.environ:
        raise MetaflowException(
            "All outerbounds app commands can only be run from a workstation."
        )

    # Every workstation has this environment variable set.
    workstation_id = os.environ["WORKSTATION_ID"]

    try:
        try:
            conf = init_config()
            metaflow_token = conf["METAFLOW_SERVICE_AUTH_KEY"]
            api_url = conf["OBP_API_SERVER"]

            workstations_response = requests.get(
                f"https://{api_url}/v1/workstations",
                headers={"x-api-key": metaflow_token},
            )
            workstations_response.raise_for_status()
        except:
            raise MetaflowException("Failed to list workstations!")

        workstations_json = workstations_response.json()["workstations"]
        for workstation in workstations_json:
            if workstation["instance_id"] == os.environ["WORKSTATION_ID"]:
                if "named_ports" in workstation["spec"]:
                    try:
                        ensure_app_start_request_is_valid(
                            workstation["spec"]["named_ports"], port, name
                        )
                    except ValueError as e:
                        raise MetaflowException(str(e))

                    for named_port in workstation["spec"]["named_ports"]:
                        if int(named_port["port"]) == port:
                            if named_port["enabled"] and named_port["name"] == name:
                                print(f"App {name} started on port {port}!")
                                print(
                                    f"Browser URL: https://{api_url.replace('api', 'ui')}/apps/{os.environ['WORKSTATION_ID']}/{name}/"
                                )
                                print(
                                    f"API URL: https://{api_url}/apps/{os.environ['WORKSTATION_ID']}/{name}/"
                                )
                                return
                            else:
                                try:
                                    response = requests.put(
                                        f"https://{api_url}/v1/workstations/update/{workstation_id}/namedports",
                                        headers={"x-api-key": metaflow_token},
                                        json={
                                            "port": port,
                                            "name": name,
                                            "enabled": True,
                                        },
                                    )

                                    response.raise_for_status()
                                    poll_success = wait_for_app_port_to_be_accessible(
                                        api_url,
                                        metaflow_token,
                                        workstation_id,
                                        name,
                                        APP_READY_POLL_TIMEOUT_SECONDS,
                                    )
                                    if poll_success:
                                        print(f"App {name} started on port {port}!")
                                        print(
                                            f"Browser URL: https://{api_url.replace('api', 'ui')}/apps/{os.environ['WORKSTATION_ID']}/{name}/"
                                        )
                                        print(
                                            f"API URL: https://{api_url}/apps/{os.environ['WORKSTATION_ID']}/{name}/"
                                        )
                                    else:
                                        raise MetaflowException(
                                            f"The app could not be deployed in {APP_READY_POLL_TIMEOUT_SECONDS / 60} minutes. Please try again later."
                                        )
                                except Exception:
                                    raise MetaflowException(
                                        f"Failed to start app {name} on port {port}!"
                                    )
    except Exception as e:
        raise MetaflowException(f"Failed to start app {name} on port {port}!")


def ensure_app_start_request_is_valid(existing_named_ports, port: int, name: str):
    """
    Ensures that the port number is available on the workstation and that an app of
    the same name is not already opened on a different port.

    Args:
        existing_named_ports: A list of named ports on the workstation.
        port: The port number to check.
        name: The name of the app to check.
    """
    existing_apps_by_port = {np["port"]: np for np in existing_named_ports}

    if port not in existing_apps_by_port:
        raise MetaflowException(f"Port {port} not found on workstation")

    for existing_named_port in existing_named_ports:
        if (
            name == existing_named_port["name"]
            and existing_named_port["port"] != port
            and existing_named_port["enabled"]
        ):
            raise MetaflowException(
                f"App with name '{name}' is already deployed on port {existing_named_port['port']}"
            )


def wait_for_app_port_to_be_accessible(
    api_url, metaflow_token, workstation_id, app_name, poll_timeout_seconds
) -> bool:
    """
    Waits for the app to be ready by polling the workstation status.
    """
    num_retries_per_request = 3
    start_time = time.time()
    retry_delay = 1.0
    poll_interval = 10
    wait_message = f"App {app_name} is currently being deployed..."
    while time.time() - start_time < poll_timeout_seconds:
        for _ in range(num_retries_per_request):
            try:
                workstations_response = requests.get(
                    f"https://{api_url}/v1/workstations",
                    headers={"x-api-key": metaflow_token},
                )
                workstations_response.raise_for_status()
                if is_app_ready(workstations_response.json(), workstation_id, app_name):
                    print(wait_message)
                    time.sleep(APP_READY_EXTRA_BUFFER_SECONDS)
                    return True
                else:
                    print(wait_message)
                    time.sleep(poll_interval)
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):
                time.sleep(retry_delay)
                retry_delay *= 2  # Double the delay for the next attempt
                retry_delay += random.uniform(0, 1)  # Add jitter
                retry_delay = min(retry_delay, 10)
    return False


def is_app_ready(response_json: dict, workstation_id: str, app_name: str) -> bool:
    """Checks if the app is ready in the given workstation's response."""
    workstations = response_json.get("workstations", [])
    for workstation in workstations:
        if workstation.get("instance_id") == workstation_id:
            hosted_apps = workstation.get("status", {}).get("hosted_apps", [])
            for hosted_app in hosted_apps:
                if hosted_app.get("name") == app_name:
                    return bool(hosted_app.get("ready"))
    return False



from metaflow.user_configs.config_decorators import (
    MutableFlow,
    MutableStep,
    CustomFlowDecorator,
)
import os

NEBIUS_ENDPOINT_URL = "https://storage.eu-north1.nebius.cloud:443"


class nebius_checkpoints(CustomFlowDecorator):

    """

    This decorator is used for setting the nebius's S3 compatible object store as the artifact store for
    checkpoints/models created by the flow.

    Parameters
    ----------
    secrets: list
        A list of secrets to be added to the step. These secrets should contain any secrets that are required globally and the secret
        for the nebius object store. The secret should contain the following keys:
        - NEBIUS_ACCESS_KEY
        - NEBIUS_SECRET_KEY

    bucket_path: str
        The path to the bucket to store the checkpoints/models.

    endpoint_url: str
        The endpoint url for the nebius object store. Defaults to `https://storage.eu-north1.nebius.cloud:443`

    Usage
    -----
    ```python
    from metaflow import checkpoint, step, FlowSpec, nebius_checkpoints

    @nebius_checkpoints(secrets=[], bucket_path=None)
    class MyFlow(FlowSpec):
        @checkpoint
        @step
        def start(self):
            # Saves the checkpoint in the nebius object store
            current.checkpoint.save("./foo.txt")

        @step
        def end(self):
            pass
    ```
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self, *args, **kwargs):
        self.bucket_path = kwargs.get("bucket_path", None)

        self.secrets = kwargs.get("secrets", [])
        if self.bucket_path is None:
            raise ValueError(
                "`bucket_path` keyword argument is required for the coreweave_datastore"
            )
        if not self.bucket_path.startswith("s3://"):
            raise ValueError(
                "`bucket_path` must start with `s3://` for the coreweave_datastore"
            )

        self.nebius_endpoint_url = kwargs.get("endpoint_url", NEBIUS_ENDPOINT_URL)
        if self.secrets is None:
            raise ValueError(
                "`secrets` keyword argument is required for the coreweave_datastore"
            )

    def evaluate(self, mutable_flow: MutableFlow) -> None:
        from metaflow import (
            checkpoint,
            model,
            huggingface_hub,
            secrets,
            with_artifact_store,
        )

        def _add_secrets(step: MutableStep) -> None:
            decos_to_add = []
            swapping_decos = {
                "huggingface_hub": huggingface_hub,
                "model": model,
                "checkpoint": checkpoint,
            }
            already_has_secrets = False
            secrets_present_in_deco = []
            for d in step.decorators:
                if d.name in swapping_decos:
                    decos_to_add.append((d.name, d.attributes))
                elif d.name == "secrets":
                    already_has_secrets = True
                    secrets_present_in_deco.extend(d.attributes["sources"])

            # If the step aleady has secrets then take all the sources in
            # the secrets and add the addtional secrets to the existing secrets
            secrets_to_add = self.secrets
            if already_has_secrets:
                secrets_to_add.extend(secrets_present_in_deco)

            secrets_to_add = list(set(secrets_to_add))

            if len(decos_to_add) == 0:
                if already_has_secrets:
                    step.remove_decorator("secrets")

                step.add_decorator(
                    secrets,
                    sources=secrets_to_add,
                )
                return

            for d, _ in decos_to_add:
                step.remove_decorator(d)

            step.add_decorator(
                secrets,
                sources=secrets_to_add,
            )
            for d, attrs in decos_to_add:
                _deco_to_add = swapping_decos[d]
                step.add_decorator(_deco_to_add, **attrs)

        def _nebius_config():
            return {
                "root": self.bucket_path,
                "client_params": {
                    "aws_access_key_id": os.environ.get("NEBIUS_ACCESS_KEY"),
                    "aws_secret_access_key": os.environ.get("NEBIUS_SECRET_KEY"),
                    "endpoint_url": self.nebius_endpoint_url,
                },
            }

        mutable_flow.add_decorator(
            with_artifact_store,
            type="s3",
            config=_nebius_config,
        )

        for step_name, step in mutable_flow.steps:
            _add_secrets(step)
